RFA: Fix uint128_t range checking in VRP

Richard Sandiford rdsandiford@googlemail.com
Sun Dec 23 16:49:00 GMT 2012


The minimum uint128_t value is an all-zeros double_int and the maximum
value is an all-ones double_int.  Truncating these values to int and
then sign-extending them gives the same all-zeros and all-ones values,
so tree-vrp.c:range_fits_type_p concludes that uint128_t fits in an int.

This showed up as a miscompilation of __fractutasf on mips64-linux-gnu,
which in turn showed up in convert-float-4.c.

There was already code to handle this problem when changing the sign
and keeping the precision, but I think the rule applies regardless
of precision.

Tested on x86_64-linux-gnu and mips64-linux-gnu.  OK to install?

Richard


gcc/
	* tree-vrp.c (range_fits_type_p): Require the MSB of the double_int
	to be clear for sign changes.

gcc/testsuite/
	* gcc.dg/torture/fp-int-convert-2.c: New test.

Index: gcc/testsuite/gcc.dg/torture/fp-int-convert-2.c
===================================================================
--- /dev/null	2012-12-03 19:06:31.446621561 +0000
+++ gcc/testsuite/gcc.dg/torture/fp-int-convert-2.c	2012-12-23 13:46:35.449077882 +0000
@@ -0,0 +1,18 @@
+/* { dg-do run } */
+/* { dg-require-effective-target int128 } */
+
+extern void abort (void);
+
+float __attribute__((noinline))
+f (__uint128_t x)
+{
+  return x + 1;
+}
+
+int
+main (void)
+{
+  if (f (0xffffffffu) == 0)
+    abort ();
+  return 0;
+}



More information about the Gcc-patches mailing list
Index: gcc/tree-vrp.c
===================================================================
--- gcc/tree-vrp.c	2012-12-23 11:50:39.046419271 +0000
+++ gcc/tree-vrp.c	2012-12-23 11:51:17.814677008 +0000
@@ -8766,9 +8766,11 @@ range_fits_type_p (value_range_t *vr, un
       || TREE_CODE (vr->max) != INTEGER_CST)
     return false;
 
-  /* For precision-preserving sign-changes the MSB of the double-int
-     has to be clear.  */
-  if (src_precision == precision
+  /* For sign changes, the MSB of the double_int has to be clear.
+     An unsigned value with its MSB set cannot be represented by
+     a signed double_int, while a negative value cannot be represented
+     by an unsigned double_int.  */
+  if (TYPE_UNSIGNED (src_type) != unsigned_p
       && (TREE_INT_CST_HIGH (vr->min) | TREE_INT_CST_HIGH (vr->max)) < 0)
     return false;
 